package nl.tudelft.ewi.es.mbdlistener.input;

import java.io.InputStream;
import java.util.TreeSet;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import nl.tudelft.ewi.es.mbdlistener.diagnoser.Conflict;
import nl.tudelft.ewi.es.mbdlistener.diagnoser.Environment;
import nl.tudelft.ewi.es.mbdlistener.diagnoser.LinkEnvironment;
import nl.tudelft.ewi.es.mbdlistener.diagnoser.NodeEnvironment;
import nl.tudelft.ewi.es.mbdlistener.diagnoser.SensorEnvironment;
import nl.tudelft.ewi.es.mbdlistener.events.ConflictCancellationEvent;
import nl.tudelft.ewi.es.mbdlistener.events.ConflictEvent;
import nl.tudelft.ewi.es.mbdlistener.events.EndEvent;
import nl.tudelft.ewi.es.mbdlistener.events.EventList;
import nl.tudelft.ewi.es.mbdlistener.events.MessageEvent;
import nl.tudelft.ewi.es.mbdlistener.events.NodeDeathEvent;
import nl.tudelft.ewi.es.mbdlistener.events.SensorDeathEvent;
import nl.tudelft.ewi.es.mbdlistener.events.StartEvent;

public class XmlInput extends ListenerInput {

	private final InputStream inputStream;
	private long timeDivisor;
	private static final QName timeStampName = new QName("timestamp");
	private static final QName serialSourceName = new QName("serial_source");
	private static final QName senderName = new QName("sender");
	private static final QName receiverName = new QName("receiver");
	private static final QName nodeIdName = new QName("nodeId");
	private static final QName idName = new QName("id");
	private static final QName sourceName = new QName("source");

	public XmlInput(EventList eventList, InputStream inputStream, boolean testbedTime) {

		super(eventList);
		this.inputStream = inputStream;
		if (testbedTime)
			this.timeDivisor = 1000000l;
		else
			this.timeDivisor = 10000000000l;
	}

	@Override
	public void start() {
		XMLInputFactory inputFactory = XMLInputFactory.newInstance();
		XMLEventReader eventReader;
		try {
			eventReader = inputFactory.createXMLEventReader(inputStream);
		} catch (XMLStreamException e) {
			throw new RuntimeException(e);
		}

		Conflict currentConflict = null;
		double lastTimeStamp = 0.0;

		while(eventReader.hasNext()){
			XMLEvent event;
			try {
				event = eventReader.nextEvent();
			} catch (XMLStreamException e) {
				throw new RuntimeException(e);
			}

			if (event.isStartElement()) {
				StartElement startElement = event.asStartElement();
				if (startElement.getName().getLocalPart() == ("run")) {
					eventList.clear();
					eventList.add(new StartEvent(0.0));
				} else {
					if (startElement.getName().getLocalPart() == ("conflict_message")) {
						lastTimeStamp = getTimeStamp(startElement);
						eventList.add(new MessageEvent(lastTimeStamp, getSerialSouce(startElement)));
					}
					else if (startElement.getName().getLocalPart() == ("env_link")) {
						currentConflict.getConflict().add(new LinkEnvironment(getSender(startElement), getReceiver(startElement)));
					} 
					else if (startElement.getName().getLocalPart() == ("env_node")) {
						currentConflict.getConflict().add(new NodeEnvironment(getNodeId(startElement)));
					}
					else if (startElement.getName().getLocalPart() == ("env_sensor")) {
						currentConflict.getConflict().add(new SensorEnvironment(getNodeId(startElement)));
					} else if (startElement.getName().getLocalPart() == ("conflict")) {
						lastTimeStamp = getTimeStamp(startElement);
						currentConflict = new Conflict(new TreeSet<Environment>(), getId(startElement), getSource(startElement), lastTimeStamp);
					}
					else if (startElement.getName().getLocalPart() == ("conflict_cancelled")) {
						lastTimeStamp = getTimeStamp(startElement);
						eventList.add(new ConflictCancellationEvent(lastTimeStamp, getSource(startElement), getId(startElement)));
					} 
					else if (startElement.getName().getLocalPart() == ("node_death_message")) {
						lastTimeStamp = getTimeStamp(startElement);
						eventList.add(new NodeDeathEvent(lastTimeStamp, getSerialSouce(startElement)));
					} 
					else if (startElement.getName().getLocalPart() == ("sensor_death_message")) {
						lastTimeStamp = getTimeStamp(startElement);
						eventList.add(new SensorDeathEvent(lastTimeStamp, getSerialSouce(startElement)));
					}
					else {
						System.err.println("Unknown XML tags: " + event.asStartElement().getName().getLocalPart());
					}
				}
			}
			if (event.isEndElement()) {
				if (event.asEndElement().getName().getLocalPart() == ("run")) {
					eventList.add(new EndEvent(lastTimeStamp+1));
				}
				else if (event.asEndElement().getName().getLocalPart() == ("conflict")) {
					if (currentConflict == null)
						throw new RuntimeException("Conflict not initialised, but at end of conflict tag");
					eventList.add(new ConflictEvent(currentConflict.getTimestamp(), currentConflict));
				}            	
			}
		}
	}

	private Integer getSource(StartElement startElement) {
		return Integer.valueOf(startElement.getAttributeByName(sourceName).getValue());
	}

	private Integer getId(StartElement startElement) {
		return Integer.valueOf(startElement.getAttributeByName(idName).getValue());
	}

	private Integer getNodeId(StartElement startElement) {
		return Integer.valueOf(startElement.getAttributeByName(nodeIdName).getValue());
	}

	private Integer getReceiver(StartElement startElement) {
		return Integer.valueOf(startElement.getAttributeByName(receiverName).getValue());
	}

	private Integer getSender(StartElement startElement) {
		return Integer.valueOf(startElement.getAttributeByName(senderName).getValue());
	}

	private Integer getSerialSouce(StartElement startElement) {
		return Integer.valueOf(startElement.getAttributeByName(serialSourceName).getValue());
	}

	private Double getTimeStamp(StartElement startElement) {
		Long timeStamp = Long.valueOf(startElement.getAttributeByName(timeStampName).getValue());
		return (double)timeStamp/timeDivisor;
	}
}
